// Created by 超悟空 on 2018/9/20.
// Version 1.0 2018/9/20
// Since 1.0 2018/9/20

import 'dart:async';
import 'package:meta/meta.dart';

import 'communication.dart';
import '_print.dart';

/// [Work]返回的数据包装类
///
/// 包含响应的全部数据，[T]类型的业务数据实例，[success]表示成功失败，
/// [message]服务响应的消息，http响应码[code]，请求传入的参数[params],
/// 服务正真有用的数据对象[result]。
class WorkData<T> {
  /// 本次服务成功失败标志
  bool _success = false;

  /// 服务响应消息
  String _message;

  /// 任务传入参数列表
  List _params;

  /// 任务结果数据
  T _result;

  /// 任务取消标志
  bool _cancel = false;

  /// 用于网络请求使用的参数
  Options _options;

  /// http响应数据
  ///
  /// 在[Work._onParseResponse]生命周期阶段开始出现
  Response _response;

  /// 在一次[Work]执行生命周期中传递的自定义数据
  ///
  /// 通常在[Work]的某个生命周期方法中创建，以便在另一个生命周期中使用。
  /// 此额外属性并不参与网络请求
  dynamic extra;

  /// 判断本次服务请求是否成功(用户接口协议约定的请求结果，并非http的请求结果，但是http请求失败时该值总是返回false)
  bool get success => _success;

  /// 获取本次请求返回的结果消息(用户接口协议中约定的消息或者根据规则生成的本地信息，并非http响应消息）
  String get message => _message;

  /// 获取任务传入的参数列表
  List get params => _params;

  /// 获取处理完成的最终结果数据(用户接口协议中定义的有效数据转化成的本地类)
  T get result => _result;

  /// 任务是否被取消
  bool get cancel => _cancel;

  /// 用于网络请求使用的参数
  Options get options => _options;

  /// http响应数据
  ///
  /// 在[Work._onParseResponse]生命周期阶段开始出现
  Response get response => _response;
}

/// 网络请求工具
const Communication _communication = Communication();

/// 任务流程的基本模型
///
/// [D]为关联的接口结果数据类型，[T]为接口响应包装类型[WorkData]
abstract class Work<D, T extends WorkData<D>> {
  /// 日志标签
  String _tag;

  /// 日志标签
  String get tag => _tag ?? _createTag();

  /// 创建日志标签
  String _createTag() {
    _tag = '$runtimeType@${hashCode.toRadixString(16)}';
    return _tag;
  }

  /// 任务取消状态标签
  bool _cancelMark = false;

  /// 取消请求工具
  final CancelToken _cancelToken = CancelToken();

  /// 启动的任务计数器
  int _counter = 0;

  /// 队尾的任务
  Future<T> _lastFuture;

  /// 启动任务
  ///
  /// * [params]为任务参数列表，[retry]为重试次数。
  /// * [onSendProgress]为数据发送进度监听器，[onReceiveProgress]为数据接收进度监听器，
  /// 在[HttpMethod.download]请求中为下载进度，在其他类型请求中为上传/发送进度。
  /// * 同一个[Work]可以多次启动任务，多次启动的任务会顺序执行。
  Future<T> start([
    List params = const [],
    int retry = 0,
    OnProgress onSendProgress,
    OnProgress onReceiveProgress,
  ]) async {
    final counter = ++_counter;

    log(tag, "No.$counter work start");

    final completer = Completer<T>();
    final lastFuture = _lastFuture;
    _lastFuture = completer.future;

    if (counter > 1) {
      await lastFuture;
    } else {
      _cancelMark = false;
    }

    // 创建数据模型
    final data = onCreateWorkData();

    data._params = params;

    // 是否继续执行
    var next = true;

    if (!_cancelMark) {
      // 执行前导任务
      next = await _onStartWork(data);
    }

    if (!_cancelMark && next) {
      // 构建http请求选项
      data._options = await _onCreateOptions(
        params,
        retry,
        onSendProgress,
        onReceiveProgress,
      );
      // 执行核心任务
      await _onDoWork(data);
    }

    if (!_cancelMark) {
      // 执行后继任务
      await _onStopWork(data);
    }

    if (!_cancelMark) {
      // 最后执行
      log(tag, "onFinish invoke");
      try {
        await onFinish(data);
      } catch (e) {
        log(tag, "onFinish failed", e);
      }
    }

    if (_cancelMark) {
      // 任务被取消
      log(tag, "onCanceled invoked");
      try {
        await onCanceled(data);
      } catch (e) {
        log(tag, "onCanceled failed", e);
      }
    }

    log(tag, "No.$counter work end");

    data._cancel = _cancelMark;

    if (--_counter == 0) {
      _lastFuture = null;
    }

    completer.complete(data);

    return completer.future;
  }

  /// 创建数据模型对象的实例
  @protected
  T onCreateWorkData();

  /// 任务启动前置方法
  ///
  /// [data]为任务将要返回的数据模型，返回true表示继续执行
  @protected
  @mustCallSuper
  Future<bool> _onStartWork(T data) async {
    // 校验参数
    if (!await onCheckParams(data.params)) {
      // 数据异常
      log(tag, "onStartWork params error");
      // 执行异常回调
      data._message = await onParamsError(data.params);
      return false;
    }

    return true;
  }

  /// 构建请求选项参数
  Future<Options> _onCreateOptions(
    List params,
    int retry,
    OnProgress onSendProgress,
    OnProgress onReceiveProgress,
  ) async {
    log(tag, "_onCreateOptions");

    final data = Map<String, dynamic>();
    await onPreFillParams(data, params);
    await onFillParams(data, params);

    final options = Options()
      ..retry = retry
      ..onSendProgress = onSendProgress
      ..onReceiveProgress = onReceiveProgress
      ..method = httpMethod
      ..headers = await onHeaders(params)
      ..params = await onPostFillParams(data, params) ?? data
      ..url = onUrl(params);

    await onConfigOptions(options, params);

    options.cancelToken = _cancelToken;

    return options;
  }

  /// 核心任务执行
  ///
  /// 此处为真正启动http请求的方法
  Future<void> _onDoWork(T data) async {
    if (_cancelMark) {
      return;
    }

    // 创建网络请求工具
    final communication = await onInterceptCreateCommunication(data) ?? _communication;

    if (_cancelMark) {
      return;
    }

    await onWillRequest(data);

    if (_cancelMark) {
      return;
    }

    data._response = await communication.request(tag, data.options);

    if (_cancelMark) {
      return;
    }

    await _onParseResponse(data);
  }

  /// 任务完成后置方法
  @mustCallSuper
  @protected
  Future<void> _onStopWork(T data) async {
    log(tag, "onStopWork invoked");

    if (!_cancelMark) {
      try {
        // 不同结果的后继执行
        if (data.success) {
          log(tag, "onSuccess invoke");
          await onSuccess(data);
        } else {
          log(tag, "onFailed invoke");
          await onFailed(data);
        }
      } catch (e) {
        log(tag, 'onStopWork failed', e);
      }
    }
  }

  /// 最后执行的一个方法
  ///
  /// 即设置请求结果和返回数据之后，并且在回调任务发送后才执行此函数
  @protected
  FutureOr<void> onFinish(T data) {}

  /// 任务被取消时调用
  @protected
  FutureOr<void> onCanceled(T data) {}

  /// 参数合法性检测
  ///
  /// * 用于检测传入参数[params]是否合法，需要子类重写检测规则。
  /// * 检测成功任务才会被正常执行，如果检测失败则[onParamsError]会被调用，
  /// 且后续网络请求任务不再执行，任务任然可以正常返回并执行生命周期[onFailed]，[onFinish]。
  /// * 参数合法返回true，非法返回false。
  @protected
  FutureOr<bool> onCheckParams(List params) => true;

  /// 参数检测不合法时调用
  ///
  /// * [onCheckParams]返回false时被调用，且后续网络请求任务不再执行，
  /// 但是任务任然可以正常返回并执行生命周期[onFailed]，[onFinish]。
  /// * 返回错误消息内容，将会设置给[WorkData.message]
  @protected
  FutureOr<String> onParamsError(List params) => null;

  /// 填充请求所需的前置参数
  ///
  /// * 适合填充项目中所有接口必须传递的固定参数（通过项目中实现的定制[Work]基类完成）
  /// * [data]为请求参数集（http请求要发送的参数），[params]为任务传入的参数列表
  @protected
  FutureOr<void> onPreFillParams(Map<String, dynamic> data, List params) {}

  /// 填充请求所需的参数
  ///
  /// [data]为请求参数集（http请求要发送的参数），[params]为任务传入的参数列表
  @protected
  FutureOr<void> onFillParams(Map<String, dynamic> data, List params);

  /// 填充请求所需的后置参数
  ///
  /// * 适合对参数进行签名（通过项目中实现的定制[Work]基类完成）
  /// * [data]为请求参数集（http请求要发送的参数），[params]为任务传入的参数列表
  /// * 如果需要使用其他数据类型作为请求参数，请返回新的数据集合对象，支持[Map]，[List]，[String]([ResponseType.plain])
  /// 不返回参数或返回null则继续使用[data]作为请求参数
  @protected
  FutureOr<dynamic> onPostFillParams(Map<String, dynamic> data, List params) {}

  /// 创建并填充请求头
  ///
  /// [params]为任务传入的参数
  @protected
  FutureOr<Map<String, dynamic>> onHeaders(List params) => null;

  /// 拦截创建网络请求工具
  ///
  /// 用于创建完全自定义实现的网络请求工具。
  @protected
  FutureOr<Communication> onInterceptCreateCommunication(T data) => null;

  /// 即将执行网络请求前的回调
  ///
  /// 此处可以用于做数据统计，特殊变量创建等，如果调用[cancel]则会拦截接下来的网络请求
  @protected
  FutureOr<void> onWillRequest(T data) => null;

  /// 自定义配置http请求选择项
  ///
  /// * [options]为请求将要使用的配置选项，[params]为任务参数
  /// 修改[options]的属性以定制http行为。
  /// * [options]包含[httpMethod]返回的请求方法，
  /// [onFillParams]填充的参数，
  /// [onUrl]返回的请求地址，
  /// [start]中传传递的[retry]和[onProgress]，
  /// [onHeaders]中创建的请求头，
  /// 以上属性都可以在这里被覆盖可以被覆盖。
  @protected
  FutureOr<void> onConfigOptions(Options options, List params) {}

  /// 网络请求方法
  @protected
  HttpMethod get httpMethod => HttpMethod.get;

  /// 网络请求完整地址
  ///
  /// [params]任务传入的参数
  @protected
  String onUrl(List params);

  /// 解析响应数据
  Future<void> _onParseResponse(T data) async {
    log(tag, "_onParse response parse start");

    if (data.response.success) {
      // 解析数据
      if (await _onParse(data)) {
        // 解析成功
        log(tag, "_onParseResponse result parse success onParseSuccess invoke");
        // 解析成功回调
        await onParseSuccess(data);
        if (data.success) {
          log(tag, "work success");
        } else {
          log(tag, "work failed");
        }
      } else {
        // 解析失败
        log(tag, "_onParseResponse result parse failed onParseFailed invoke");
        // 解析失败回调
        data._success = false;
        data.response.errorType = HttpErrorType.parse;
        data._message = await onParseFailed(data);
      }
    } else if (data.response.errorType == HttpErrorType.response) {
      // 网络请求失败
      log(tag, "_onParseResponse network request false onNetworkRequestFailed invoke");

      // 网络请求失败回调
      data._message = await onNetworkRequestFailed(data);
    } else {
      // 网络连接失败
      log(tag, "_onParseResponse network error onNetworkError invoke");

      // 网络错误回调
      data._message = await onNetworkError(data);
    }
  }

  /// 解析响应体，返回解析结果
  Future<bool> _onParse(T data) async {
    log(tag, "_onParse start");
    if (!await onCheckResponse(data)) {
      // 通信异常
      log(tag, "_onParse response body error");
      return false;
    }

    try {
      // 提取服务执行结果
      data._success = await onResponseResult(data);
      log(tag, "_onParse request result:${data.success}");

      if (data.success) {
        // 服务请求成功回调
        log(tag, "_onParse onRequestSuccess invoked");
        data._result = await onResponseSuccess(data);
        // 提取服务返回的消息
        data._message = await onRequestSuccessMessage(data);
      } else {
        // 服务请求失败回调
        log(tag, "_onParse onRequestFailed invoked");
        data._result = await onRequestFailed(data);
        // 提取服务返回的消息
        data._message = await onRequestFailedMessage(data);
        data.response.errorType = HttpErrorType.task;
      }
      log(tag, "_onParse request message:", data.message);

      return true;
    } catch (e) {
      log(tag, "_onParse error:", e);
      return false;
    } finally {
      log(tag, "_onParse end");
    }
  }

  /// 服务器响应数据解析成功后调用
  ///
  /// 即在[_onParse]返回true时调用
  @protected
  FutureOr<void> onParseSuccess(T data) {}

  /// 网络请求成功，服务器响应数据解析失败后调用
  ///
  /// 即在[_onParse]返回false时调用，
  /// 返回响应数据解析失败时的消息，即[WorkData.message]字段
  @protected
  FutureOr<String> onParseFailed(T data) => null;

  /// 网络连接建立成功，但是请求失败时调用
  ///
  /// 即响应码不是200，如4xx，5xx，返回网络请求失败时的消息，即[WorkData.message]字段
  @protected
  FutureOr<String> onNetworkRequestFailed(T data) => null;

  /// 网络连接建立失败时调用，即网络不可用
  ///
  /// 返回设置网络无效时的消息，即[WorkData.message]字段
  @protected
  FutureOr<String> onNetworkError(T data) => null;

  /// 检测响应结果是否符合预期（数据类型或是否包含特定字段），也可以做验签
  ///
  /// * 通常[data.response]类型是[onConfigOptions]中设置的[Options.responseType]决定的。
  /// * 在一般请求中默认为[ResponseType.json]则[data.response]为[Map]类型的json数据。
  /// * 下载请求中默认为[ResponseType.stream]则[data.response]为[Stream]。
  /// * 如果设置为[ResponseType.plain]则[data.response]为字符串。
  @protected
  FutureOr<bool> onCheckResponse(T data) => true;

  /// 提取服务执行结果
  ///
  /// * http响应成功，从接口响应的数据中提取本次业务请求真正的成功或失败结果。
  /// * 通常[data.response]类型是[onConfigOptions]中设置的[Options.responseType]决定的。
  /// * 在一般请求中默认为[ResponseType.json]则[data.response]为[Map]类型的json数据。
  /// * 下载请求中默认为[ResponseType.stream]则[data.response]为[Stream]。
  /// * 如果设置为[ResponseType.plain]则[data.response]为字符串。
  @protected
  FutureOr<bool> onResponseResult(T data);

  /// 提取服务执行成功时返回的真正有用结果数据
  ///
  /// * 在服务请求成功后调用，即[onResponseResult]返回值为true时被调用，
  /// 用于生成请求成功后的任务返回真正结果数据对象[D]。
  /// * 通常[data.response]类型是[onConfigOptions]中设置的[Options.responseType]决定的。
  /// * 在一般请求中默认为[ResponseType.json]则[data.response]为[Map]类型的json数据。
  /// * 下载请求中默认为[ResponseType.stream]则[data.response]为[Stream]。
  /// * 如果设置为[ResponseType.plain]则[data.response]为字符串。
  @protected
  FutureOr<D> onResponseSuccess(T data);

  /// 提取或设置服务返回的成功结果消息
  ///
  /// 在服务请求成功后调用，即[onResponseResult]返回值为true时被调用。
  @protected
  FutureOr<String> onRequestSuccessMessage(T data) => null;

  /// 提取或设置服务执行失败时的返回结果数据
  ///
  /// 在服务请求失败后调用，即[onResponseResult]返回值为false时被调用，
  /// 用于生成请求失败后的任务返回真正结果数据对象[D]，可能是一个默认值。
  @protected
  FutureOr<D> onRequestFailed(T data) => null;

  /// 提取或设置服务返回的失败结果消息
  ///
  /// 在服务请求失败后调用，即[onResponseResult]返回值为false时被调用。
  FutureOr<String> onRequestFailedMessage(T data) => null;

  /// 本次任务执行成功后执行
  ///
  /// 即设置请求结果和返回数据之后，并且在回调接口之前执行此函数，
  /// 该方法在[onFinish]之前被调用
  @protected
  FutureOr<void> onSuccess(T data) {}

  /// 本次任务执行失败后执行
  ///
  /// 即设置请求结果和返回数据之后，并且在回调接口之前执行此函数，
  /// 该方法在[onFinish]之前被调用
  @protected
  FutureOr<void> onFailed(T data) {}

  /// 取消正在进行的任务
  ///
  /// 如果本任务被多次启动排队执行，则会一次性取消所有排队任务和正在执行的任务
  void cancel() {
    log(tag, "cancel");
    if (_cancelMark) {
      log(tag, "work has been canceled");
      return;
    }
    if (_counter <= 0) {
      log(tag, "work not started");
      return;
    }
    _cancelMark = true;
    _cancelToken.cancel();
  }
}
